﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Core;
using HeuristicLab.Random;

namespace HeuristicLab.ExpressionGenerator {
  public abstract class ExpressionTemplate {
    protected List<Tuple<Expression, double>> arguments;
    protected readonly Func<IEnumerable<double>, double> transform;
    protected readonly string label;

    public abstract Expression Instantiate(IRandom random, bool sampleWithRepetition = false);

    protected ExpressionTemplate(string label, Func<IEnumerable<double>, double> transform) {
      this.transform = transform;
      this.label = label;
    }

    public void AddArgument(Expression expression, double priority = 1d) {
      arguments = arguments ?? new List<Tuple<Expression, double>>();
      arguments.Add(Tuple.Create(expression, priority));
    }

    public void AddArguments(IEnumerable<Expression> expressions) {
      arguments = arguments ?? new List<Tuple<Expression, double>>();
      arguments.AddRange(expressions.Select(x => Tuple.Create(x, 1d)));
    }

    public void AddArguments(IEnumerable<Expression> expressions, IEnumerable<double> priorities) {
      arguments = arguments ?? new List<Tuple<Expression, double>>();
      arguments.AddRange(expressions.Zip(priorities, Tuple.Create));
    }
  }

  public class RandomArityTemplate : ExpressionTemplate {
    private readonly IRandom arityDistribution;

    public RandomArityTemplate(string label, Func<IEnumerable<double>, double> transform, IRandom arityDistribution) : base(label, transform) {
      this.arityDistribution = arityDistribution;
    }

    public override Expression Instantiate(IRandom random, bool sampleWithRepetition = false) {
      var arity = (int)Math.Round(arityDistribution.NextDouble());
      var weights = arguments.Select(x => x.Item2);
      var args = sampleWithRepetition
        ? arguments.SampleProportional(random, arity, weights).Select(x => x.Item1)
        : arguments.SampleProportionalWithoutRepetition(random, arity, weights).Select(x => x.Item1);
      return Expression.Function(label, transform, args);
    }
  }

  public class FixedArityTemplate : ExpressionTemplate {
    private readonly int arity;
    public FixedArityTemplate(string label, Func<IEnumerable<double>, double> transform, int arity) : base(label, transform) {
      this.arity = arity;
    }

    public override Expression Instantiate(IRandom random, bool sampleWithRepetition = false) {
      var args = arguments.SampleProportional(random, arity, arguments.Select(x => x.Item2)).Select(x => x.Item1);
      return Expression.Function(label, transform, args);
    }
  }
}